;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

;;;
;;; Query worksheet displays a menu for selecting query 
;;; operations (collect, find, maximize, etc) and a menu for selecting
;;; query limits (when, unless etc). Each menu has an associated
;;; typein buffer for entering the query or limit expression. Hitting
;;; return maps the formulated query over the current selections in the
;;; current focus browser. The results are displayed in the  query
;;; browser's output view. Text in the output view is selectable.
;;;

(defconstant +query-worksheet-size+ (xy 300 257))

(defclass query-worksheet (worksheet)
  ()
  (:default-initargs 
    :window-type :document-with-zoom
    :view-size +query-worksheet-size+))

(defmethod ccl:window-size-parts ((window query-worksheet))
  ;; have to check for subviews since mcl calls this function
  ;; at make-instance time, before any of the views are there!
  (let ((size (ccl:view-size window)))
    (let ((v (ccl:view-named ':query-input window)))
      (and v (ccl:set-view-size v (xy-x size)
                                (view-height v))))
    (let ((v (ccl:view-named ':query-output window)))
      (and v 
           #+ccl-3 (ccl:set-view-size v (1+ (xy-x size))
                                      (- (1+ (xy-y size))
                                         (xy-y (ccl:view-position v))))
           #-ccl-3 (ccl:set-view-size v (- (xy-x size) +scroller+)
                                      (- (xy-y size)
                                         (xy-y (ccl:view-position v))
                                         +scroller+))))))

(defmethod worksheet-action ((view query-worksheet))
  (worksheet-action (ccl:view-named ':query-input view)))

(defmethod initialize-instance :after ((view query-worksheet) 
                                       &rest ignore)
  (declare (ignore ignore))
  (ccl:add-subviews
   view
   (make-instance 'query-input 
     :view-position (xy 0 0) :view-size (xy (view-width view) 57))
   (make-instance #+ccl-3 'ccl::scrolling-fred-view 
                  #-ccl-3 'ccl::scrolling-fred-dialog-item
     :view-position (xy 0 57)
     :view-nick-name ':query-output
     :view-size (xy 285 (- (view-height view) 57 +scroller+))))
  (set-window-grow view +query-worksheet-size+))

(defmethod initialize-worksheet ((window query-worksheet))
  (with-focus-browser (browser)
    (if (and (listing-browser? browser)
             (> (length (ccl:view-subviews window)) 0))
      (progn 
        (set-worksheet-title window "Query" browser)
        (enable-worksheet window))
      (disable-worksheet window))))

(defmethod enable-worksheet ((sheet query-worksheet))
  (ccl:map-subviews (ccl:view-named ':query-input sheet)
                    #'control-enable))

(defmethod disable-worksheet ((sheet query-worksheet))
  (ccl:map-subviews (ccl:view-named ':query-input sheet)
                    #'control-disable))

;;;
;;; input and output views for entering and displaying a Query.
;;;

(defclass query-input (gray-3d-framed-view)
  ()
   (:default-initargs 
    :view-nick-name ':query-input
    :view-size (xy 300 50)))

(defmethod worksheet-action ((input query-input))
  (let ((qt (textbuf-text (ccl:view-named ':query-expr input)))
        (lt (textbuf-text (ccl:view-named ':limit-expr input)))
        (qm (ccl:view-named ':query-menu input))
        (lm (ccl:view-named ':limit-menu input)))
    (if (balanced-form? qt)
      (if (balanced-form? lt)
        (unless (string= qt "")
          (let ((worksheet (view-worksheet input))
                (query (format nil "~A ~A"
                               (ccl:menu-item-title
                                (ccl:selected-item qm)) qt)))
            (when (control-enabled? lm)
              (unless (string= lt "")
                (setf query (concatenate
                             'string
                             (format nil "~A ~A "
                                     (ccl:menu-item-title
                                      (ccl:selected-item lm))
                                     lt)
                             query))))
            (multiple-value-bind (clause error)
                                 (parse-map-input query :only-one t)
              (if clause
                #-ccl-3
                (let* ((reference
                        (or (selections-to-reference
                             (browser-listing *browser*))
                            (makeref (object-view-object *browser*)
                                     0 ':end)))
                       (*standard-output*
                        (ccl:view-named ':query-output worksheet))
                       (buf (ccl:fred-buffer *standard-output*))
                       (pos (ccl:buffer-position buf)))
                  (mapslots reference clause nil)
                  (ccl:set-selection-range *standard-output* pos)
                  (ccl:fred-update *standard-output*))
                #+ccl-3
                (let* ((reference
                        (or (selections-to-reference
                             (browser-listing *browser*))
                            (makeref (object-view-object *browser*)
                                     0 ':end)))
                       (v (ccl:view-named ':query-output worksheet))
                       (buf (ccl:fred-buffer v))
                       (pos (ccl:buffer-position buf)))
                  (with-output-to-string (*standard-output*)
                    (fresh-line *standard-output*)
                    (mapslots reference clause nil)
                    (ccl::buffer-insert buf (slot-value *standard-output*
                                                        'ccl::my-string)
                                        pos)
                    (ccl::invalidate-view v)))
                (warn-user error)))))
        (progn
          (warn-user "Bad limit expression: ~S." lt)
          nil))
      (progn
        (warn-user "Bad query expression: ~S." qt)
        nil))))

;(defun dim-stale-output (view)
;  (let ((y (* (font-height (ccl:view-font view))
;              (ccl:lines-in-buffer (ccl:fred-buffer view))))
;        (o (ccl:pen-mode view)))
;    (when (> y 0)
;      (ccl:with-focused-view view
;        (ccl:set-pen-mode view ':patOr)
;        (ccl:fill-rect view ccl:*light-gray-pattern* (xy 0 0)
;                       (xy (view-width view) y))
;        (ccl:set-pen-mode view o)))))

(defmethod initialize-instance :after ((view query-input) &rest args)
  (declare (ignore args))
  (ccl:add-subviews
   view
   (make-instance 'invisible-default-button
     :action #'worksheet-action
     :view-nick-name ':Ok)
   (make-instance 'ccl:pop-up-menu
     :menu-items 
     (loop for op in '("Find" "Collect" "Count" "Sum" "Minimize"
                       "Maximize" "Highest" "Lowest" "Average" "Analyze")
           collect (make-instance 'ccl:menu-item :menu-item-title op))
     :view-position (xy 8 8) :view-size (xy 75 18)
     :view-nick-name ':query-menu :view-font *capella-font*)
   (make-instance 'textbuf 
     :view-position (xy 90 11)
     :view-size (xy 198 12)
     :title ""
     :action nil
     :view-nick-name ':query-expr)
   (make-instance 'ccl:pop-up-menu
     :menu-items 
     (loop for op in '("When" "Unless" "While" "Until")
           collect (make-instance 'ccl:menu-item :menu-item-title op))
     :view-position (xy 8 30) :view-size (xy 75 18)
     :view-nick-name ':limit-menu  :view-font *capella-font*)
   (make-instance 'textbuf 
     :view-position (xy 90 33)
     :view-size (xy 198 12)
     :title ""
     :action nil
     :view-nick-name ':limit-expr)))

(defmethod ccl:set-view-size :after ((view query-input) x &optional y)
  (declare (ignore y))
  (let ((b1 (ccl:view-named ':query-expr view))
        (b2 (ccl:view-named ':limit-expr view)))
    (ccl:set-view-size b1 (- (xy-x x) (xy-x (ccl:view-position b1)) 12)
                       (view-height b1))
    (ccl:set-view-size b2 (- (xy-x x) (xy-x (ccl:view-position b2)) 12)
                       (view-height b2))))

;;;
;;; Delete worksheet. 
;;;

(defclass out-worksheet (framed-worksheet)
  ()
  (:default-initargs
    :view-size (xy 205 65) ))

(defmethod initialize-instance :after ((view out-worksheet) &rest args)
  (declare (ignore args))
  (let ((xy (xy 8 8)))
    (ccl:add-subviews
     view
     (make-instance 'radio-button
       :view-position (xyinc xy 0 0)
       :title "Mute"
       :action #'update-worksheet
       :view-nick-name ':hide)
     (make-instance 'radio-button
       :view-position (xyinc xy 65 0)
       :title "Delete"
       :action #'update-worksheet
       :view-nick-name ':delete)
     (make-instance 'check-box
       :view-position (xyinc xy 128 0)
       :title "Expunge"
       :action #'update-worksheet
       :view-nick-name ':expunge)
     (make-instance 'radio-button
       :view-position (xyinc xy 0 15)
       :title "Unmute"
       :action #'update-worksheet
       :view-nick-name ':unhide)
     (make-instance 'radio-button
       :view-position (xyinc xy 65 15)
       :title "Undelete"
       :action #'update-worksheet
       :view-nick-name ':undelete)
     
     (make-instance 'default-button
       :view-position (xyinc xy 70 33)
       :view-size (xy 30 18)
       :action #'worksheet-action
       :view-nick-name ':Ok))))

(defmethod initialize-worksheet ((worksheet out-worksheet)) 
  ;; always bring worksheet up with expunge unchecked
  (with-focus-browser (browser)
    (if (listing-browser? browser)
      (progn 
        (toggle-deselect (ccl:view-named ':expunge worksheet))
        (if (any-selection? browser)
          (enable-worksheet worksheet)
          (disable-worksheet worksheet))
        (update-worksheet worksheet)
        (set-worksheet-title worksheet "Out" browser))
    (disable-worksheet worksheet))))

(defmethod enable-worksheet ((window out-worksheet))
  (ccl:map-subviews window
                    #'control-enable 'ccl:dialog-item))

(defmethod disable-worksheet ((window out-worksheet))
  (ccl:map-subviews window
                    #'control-disable 'ccl:dialog-item))

(defmethod update-worksheet ((worksheet out-worksheet))
  (let ((del (ccl:view-named ':delete worksheet))
        (exp (ccl:view-named ':expunge worksheet)))
  (if (toggle-selected? del)
    (control-enable exp)
    (control-disable exp))))

(defmethod worksheet-action ((worksheet out-worksheet))
  (with-focus-browser (browser)
    (let ((command (ccl:view-nick-name
                    (find-subview-if
                     #'(lambda (v)
                         (and (typep v 'radio-button)
                              (toggle-selected? v)))
                     worksheet))))
      (if (any-selection? browser)
        (let ((listing (browser-listing browser)))
          (when (and (eq command ':delete)
                     (toggle-selected?
                      (ccl:view-named ':expunge worksheet)))
            (setf command ':expunge))
          (listing-delete listing command)
          (when (eq command ':expunge) ;; nothing selected anymore.
            (disable-worksheet worksheet)))
        (warn-user "No selections to ~A" command)))))
;;;
;;; Edit worksheet contains command buttons and a menu that installs
;;; slot table for different classes of objects
;;;

(defclass set-worksheet (framed-worksheet)
  ()
  (:default-initargs
    :view-size (xy 305 69)))

(defclass mapping-table (slot-table)
  ())

(defmethod table-label-string (label (table mapping-table))
  ;; slot get prefixed with dollar notation.
  (concatenate 'string "$" (item-title label)))

(defmethod initialize-instance :after ((view set-worksheet) &rest args)
  (declare (ignore args))
  (let* ((size (ccl:view-size view))
         (commands (make-instance 'map-commands
                     :view-position (xy 0 0)
                     :view-size (xy (view-width view) 44)))
         (templates (make-instance 'template-menu
                      :objects (element-classes)
                      :quiet-item "Objects"
                      :action #'template-action
                      :update #'update-worksheet
                      :table-position (xy 8 60)
                      :view-position (xy 10 40)))
         (controls (make-instance 'set-controls
                     :view-position (xy 0 40))))
    (ccl:add-subviews view commands templates controls )
    ;; spread commands across window
    (ccl:set-view-size commands
                       (xy-x size)
                       (xy-y (ccl:view-size commands)))
    ;; right justify controls
    (let ((p (ccl:view-position controls))
          (s (ccl:view-size controls)))
      (ccl:set-view-position controls (- (xy-x size) 
                                         (xy-x s)
                                         (xy-x *dxy*))
                             (xy-y p)))))

(defun template-action (menu object)
  (let* ((worksheet (view-worksheet menu))
         (template (assoc object (slot-value menu 'templates)))
         (installed (ccl:view-named ':slot-table worksheet))
         (table (cdr template)))
    (when object
      (unless table
        (setf table (make-instance 'mapping-table
                      :table-color ccl:*light-gray-color*
                      :object object
                      :view-position (slot-value menu 'tablepos)
                      :view-size 
                      (xy (- (view-width worksheet) 16) 
                          0)))
        (setf (cdr template) table)))
    (install-view worksheet table installed)))

(defmethod initialize-worksheet ((worksheet set-worksheet))
  ;; try to select an appropriate template when the worksheet is selected
  (with-focus-browser (browser)
    (if (listing-browser? browser)
      (let* ((listing (browser-listing browser))
             (default (selection-class? listing %check-selections-limit%)))
        (enable-worksheet worksheet)
        (unless (toggle-selected? 
                 (find-subview (find-subview worksheet ':set-controls )
                               ':lock))
          (pop-up-menu-select
           (ccl:view-named ':template-menu worksheet)
           (if (and default (event? (class-prototype default)))
             (string (class-name default)) "Objects")))
        (set-worksheet-title worksheet "Edit" browser))
      (disable-worksheet worksheet))))

(defmethod enable-worksheet ((window set-worksheet))
  (ccl:map-subviews (ccl:view-named ':map-commands window)
                    #'control-enable 'ccl:dialog-item)
   (ccl:map-subviews (ccl:view-named ':set-controls window)
                    #'control-enable 'ccl:dialog-item))

(defmethod disable-worksheet ((window set-worksheet))
  (ccl:map-subviews (ccl:view-named ':map-commands window)
                    #'control-disable 'ccl:dialog-item)
  (ccl:map-subviews (ccl:view-named ':set-controls window)
                    #'control-disable 'ccl:dialog-item))

(defmethod update-worksheet ((worksheet set-worksheet)) 
  (let ((table (ccl:view-named ':slot-table worksheet))
        (commands (ccl:view-named ':map-commands worksheet))
        (controls (ccl:view-named ':set-controls worksheet)))
    (cond ((null table)
           (control-disable (ccl:view-named ':lock controls))
           (control-disable (ccl:view-named ':All controls))
           (control-disable (ccl:view-named ':ok controls))
           (ccl:map-subviews commands #'control-disable))
          (t
           (control-enable (ccl:view-named ':lock controls))
           (control-enable (ccl:view-named ':All controls))
           (control-enable (ccl:view-named ':ok controls))
           (ccl:map-subviews commands #'control-enable)))))

(defmethod worksheet-action ((worksheet set-worksheet))
  (with-focus-browser (browser )
    (let* ((listing (browser-listing browser))
           (table (ccl:view-named ':slot-table worksheet))
           (op (slot-value (ccl:view-named ':map-commands worksheet)
                           'command))
           (forms (get-set-forms table op)))
      (when forms
        (if (consp forms)
          (let ((fun (if (eq op ':unset)
                       #'(lambda (o c) c (unset-slots o forms))
                       #'(lambda (o c) c (set-slots o forms))))
                (lev (mapping-level))
                (mod (mapping-mode t))
                (tab (listing-view-displays listing))
                (sup (object-view-object listing)))
            (let ((wrapper
                   #'(lambda (o)
                       (map-object-aux o fun sup mod 0 lev)
                       (clear-display tab o))))
              (ccl:with-cursor ccl:*watch-cursor*
                (if (any-selection? listing)
                  (map-selected-objects wrapper listing)
                  (map-sequence-objects wrapper listing))
                (draw-changed-displays listing))))
          (warn-user "Template contains malformed data:~S" forms))))))

(defclass map-commands (ccl:view)      ; gray-3d-framed-view
  ((command :initform nil))
  (:default-initargs 
    :view-nick-name ':map-commands))

(defmethod initialize-instance :after ((view map-commands)
                                       &rest args)
  (declare (ignore args))
  (let ((commands (list
                   (make-instance 'radio-button
                     :title "Set" 
                     :action #'set-command-action
                     :view-nick-name ':Set
                     :selected? t)
                   (make-instance 'radio-button
                     :title "Unset" 
                     :action #'set-command-action
                     :view-nick-name ':Unset)
                   (make-instance 'radio-button
                     :title "Scale" 
                     :action #'set-command-action
                     :view-nick-name ':Scale)
                   (make-instance 'radio-button
                     :title "Increment" 
                     :action #'set-command-action
                     :view-nick-name ':Increment)
                   (make-instance 'radio-button
                     :title "Transpose" 
                     :action #'set-command-action
                     :view-nick-name ':Transpose)
                   (make-instance 'radio-button
                     :title "Invert" 
                     :action #'set-command-action
                     :view-nick-name ':Invert)
                   (make-instance 'radio-button
                     :title "Round" 
                     :action #'set-command-action
                     :view-nick-name ':Round)
                   (make-instance 'radio-button
                     :title "Quantize" 
                     :action #'set-command-action
                     :view-nick-name ':Quantize))))
    (setf (slot-value view 'command) :Set)
    (layout-set-commands commands (view-width view))
    (apply #'ccl:add-subviews view commands)
    view))

(defmethod ccl:set-view-size :after ((view map-commands) h &optional v)
  (layout-set-commands (ccl:view-subviews view) (if v h (xy-x h))))

(defun layout-set-commands (commands width )
  (let* ((len (length commands))
         (wid (floor (- width 16) (floor len 2))))
    (loop for x = 8 then (+ wid x)
          for i below len by 2
          for v1 = (elt commands i)
          for v2 = (elt commands (1+ i))
          do 
          (ccl:set-view-position v1 x 8)
          (ccl:set-view-position v2 x 24))))

(defun set-command-action (button)
  (let ((name (ccl:view-nick-name button))
        (last (slot-value (ccl:view-container button) 'command)))
    (unless (eq name last)
      (let ((table (ccl:view-named ':slot-table (view-browser button))))
        (when table (table-clear table))
        (setf (slot-value (ccl:view-container button) 'command)
              name)))))

;;;
;;; set-control field hold clear cancel and quit buttons, and an
;;; All box that toggles between showing all slots or parameters.
;;;

(defclass set-controls (ccl:view)
  ()
  (:default-initargs
    :view-nick-name ':set-controls
    :view-size (xy 185 15)))

(defmethod initialize-instance :after ((view set-controls)
                                       &rest args)
  (declare (ignore args))
  (ccl:add-subviews 
   view
   (make-instance 'check-box
     :view-position (xy 0 0)
     :title "Lock"
     :action #'(lambda (c) c)
     :view-nick-name ':lock)
   (make-instance 'check-box
     :view-position (xy 45 0)
     :title "Inspect"
     :action #'(lambda  (box) 
                 (let ((table (ccl:view-named ':slot-table 
                                              (view-worksheet box))))
                   (when table
                     (inspect-slots table
                                    (toggle-selected? box)))))
     :view-nick-name ':all)
   (make-instance 'button
     :view-position (xy 105 1)
     :view-size (xy 35 12)
     :title "Clear"
     :action #'(lambda (b) 
                 (let ((tab (ccl:view-named ':slot-table (view-worksheet b))))
                   (and tab (table-clear tab))))
     :view-nick-name ':clear)
   (make-instance 'default-button
     :view-position (xy 145 1)
     :view-size (xy 35 12)
     :default? nil
     :action #'worksheet-action
     :view-nick-name ':Ok)))

(defun get-set-forms (table op)
  ;; return list of forms, nil or error flag
  (let ((forms '()))
    (map-table
     #'(lambda (e)
         (when (and (table-entry-visible? e)
                    (not (table-entry-readonly? e)))
           (let ((s (textbuf-text (table-entry-editor e))))
             (if (eq op ':Unset)
               (when (string= s "") 
                 (push (table-entry-id e)
                       forms))
               (unless (string= s "") 
                 (if (balanced-form? s)
                   (progn
                     (push ;(walk-map-expr (read-from-string s) op nil nil)
                      (coerce
                       `(lambda ()
                          ,(walk-map-expr (read-from-string s) op nil nil))
                       'function)
                      forms)
                     (push (table-entry-id e)
                           forms))
                   (return-from get-set-forms s)))))))
     table)
    forms))

;;;
;;; New worksheet allows objects to be created or duplicated. After the
;;; worksheet completes the new objects are automatically selected.
;;;
                     
(defclass new-worksheet (framed-worksheet)
  ((spec :accessor new-worksheet-spec :initform nil))
  (:default-initargs
    :window-title "New"
    :view-size (xy 204 70)))

(defmethod initialize-instance :after ((view new-worksheet) &rest args)
  (declare (ignore args))
  (let ((xy (xy 8 8)))
    (ccl:add-subviews
     view
     (make-instance 'radio-button
      :view-position xy 
      :title "Duplicate"
      :action #'update-worksheet
      :view-nick-name ':Duplicate)
     (make-instance 'radio-button
      :view-position (xyinc xy 0 15) 
      :title "Specify"
      :action #'update-worksheet
      :view-nick-name ':New)
     (make-new-menu :view-position (xyinc xy 0 35)
                    :view-nick-name ':class-menu )
     (make-instance 'label 
       :view-position (xyinc xy 110 0)
       :title "Copies:")
     (make-instance 'textbuf 
       :view-position (xyinc xy 155 2)
       :view-size (xy 27 12)
       :title "1"
       :action nil
       :view-nick-name ':Copies)
     (make-instance 'default-button 
       :default? nil
       :view-position (xyinc xy 140 35) 
       :view-size (xy 45 14)
       :action #'worksheet-action
       :view-nick-name ':Ok))))

(defun make-new-menu (&rest args)
  (let* ((algotypes '(algorithm generator))
         (events (element-classes))
         menu)
    (labels ((myaction (thing)
               #'(lambda () 
                   (let ((w (ccl:view-window menu)))
                     (setf (new-worksheet-spec w) thing)
                     (update-worksheet w))))
             (makeone (thing &optional title)
               (make-instance 'ccl:menu-item
                 :menu-item-title (or title (object-namestring thing))
                 :menu-item-action (myaction thing))))
      (setf menu
            (apply 
             #'make-instance 'ccl:pop-up-menu
             :view-font (italic-face *capella-font*)
             :menu-items
             (nconc
              (list 
               (make-instance 'ccl:menu-item 
                 :menu-item-title "Objects" :disabled t
                 :style :italic ))
              (mapcar #'makeone events)
              (list (make-instance 'ccl:menu-item 
                      :menu-item-title "Containers" :disabled t
                      :style :italic ))
              (mapcar #'makeone (loop for c in (container-classes)
                                      unless (member (class-name c) 
                                                     algotypes)
                                      collect c))
              (loop for a in algotypes
                    collect
                    (let ((type a) subitems submenu)
                      (setf subitems 
                            (loop for e in events
                                  collect 
                                  (make-instance 'ccl:menu-item
                                    :menu-item-title (object-namestring e)
                                    :menu-item-action
                                    (let ((c e) (n (object-namestring e)))
                                      #'(lambda () 
                                          (let ((w (ccl:view-window menu)))
                                            (ccl:set-menu-title 
                                             submenu
                                             (format nil "~:[Gen~;Algo~]/~A"
                                                     (eq type 'algorithm) n))
                                            (setf (new-worksheet-spec w) 
                                                  #'(lambda ()
                                                      (make-object type :class c)))
                                            (update-worksheet w))) ))))
                      (setf submenu
                            (make-instance 'ccl:menu
                              :menu-title (string-capitalize type)
                              :menu-items subitems))
                      submenu))
              (list (makeone #'(lambda () (make-object 'algorithm))
                             "Mute")))
             args))
      menu)))
               
(defmethod disable-worksheet ((window new-worksheet))
  (ccl:map-subviews window #'control-disable 'ccl:dialog-item))

(defmethod enable-worksheet ((window new-worksheet))
  (ccl:map-subviews window #'control-enable 'ccl:dialog-item))

(defmethod initialize-worksheet ((worksheet new-worksheet)) 
  (with-focus-browser (browser )
    (if (listing-browser? browser)
      (let ((new (ccl:view-named ':New worksheet))
            (dup (ccl:view-named ':Duplicate worksheet))
            (menu (ccl:view-named ':class-menu worksheet))
            (copy (ccl:view-named ':copies worksheet))
            (obj (object-view-object browser)))
        (enable-worksheet worksheet)
        ;; always default to 1 copy.
        (ccl:set-dialog-item-text copy "1")
        
        ;; top-level is treated as a special case. only container
        ;; types can be created.
        (loop with top? = (eq obj .top-level.)
              and events = (element-classes)
              for i in (ccl:menu-items menu 'object-menu-item)
              when (member (object-menu-item-object i) events)
              do (if top? (ccl:menu-item-disable i)
                     (ccl:menu-item-enable i)))
        (let ((class (selection-class? browser %check-selections-limit%))
              item)
          (when (and class (not (typep (class-prototype class) 
                                       'algorithm)))
            (setf item (string (class-name class))))
          (unless item (setf (new-worksheet-spec worksheet) nil))
          (pop-up-menu-select menu (or item "Objects")))
        (ccl:menu-update menu)
        ;; enable edit choices based on number of selections. no
        ;; selections placed new objects at end of listing. 1 selection
        ;; allows either new or duplicate. else only duplicate allowed
        (cond ((no-selection? browser)
               (control-disable dup)
               (control-enable new)
               (toggle-select new))
              ((single-selection? browser)
               (control-enable dup)
               (control-enable new)
               (control-enable menu))
              (t
               (control-disable new)
               (control-enable dup)
               (toggle-select dup)))
        (set-worksheet-title worksheet "New" browser)
        (update-worksheet worksheet))
      (disable-worksheet worksheet))))

(defmethod update-worksheet ((worksheet new-worksheet))
  (let ((menu (ccl:view-named ':class-menu worksheet))
        (new? (toggle-selected?
               (ccl:view-named ':New worksheet)))
        (ok (ccl:view-named ':ok worksheet)))
    (cond (new?
           (control-enable menu)
           (if (new-worksheet-spec worksheet)
             (control-enable ok)
             (control-disable ok)))
          (t 
           (control-disable menu)
           (control-enable ok)))))

(defmethod worksheet-action ((worksheet new-worksheet))
  (with-focus-browser (browser)
    (let ((copies (textbuf-integer
                   (ccl:view-named ':copies worksheet) 1))
          (type (if (toggle-selected?
                     (ccl:view-named ':new worksheet)) 
                  (new-worksheet-spec worksheet)
                  nil))
          (listing (browser-listing browser)))
      (when type
        (unless (functionp type)
          (setf type (class-name type))))
      (listing-new listing copies type)
      (update-browser browser))))

;;;
;;; preferences
;;;

(defclass preferences (framed-worksheet)
  ((current :initform nil :accessor preferences-current))
  (:default-initargs 
    :view-size (xy 240 126)
    :window-title "Capella Preferences"))

(defmethod initialize-instance :after ((view preferences) &rest args)
  (declare (ignore args))
  (flet ((font-names (fonts key)
           (loop for n in fonts collect
                 (let ((n n))
                   (make-instance 'ccl:menu-item
                     :menu-item-title n
                     :menu-item-action
                     #'(lambda (&aux (p (preferences-current view)))
                         (setf (first (getf p key)) n))))))
         (font-sizes (sizes key)
           (loop for n in sizes collect
                 (let ((n n))
                   (make-instance 'ccl:menu-item
                     :menu-item-title (prin1-to-string n)
                     :menu-item-action
                     #'(lambda (&aux (p (preferences-current view)))
                         (setf (second (getf p key)) n)))))))
    (let ((fonts ccl:*font-list*)
          (sizes '(8 9 10 12 14 16))
          (h (* (font-height *capella-font*) 2))
          p1 p2 p3 s1 s2 s3 w)
      (setf p1 (xy 8 8))
      (setf s1 (string-extent "Background Color:" *capella-font*
                              (xy 3 3)))
      (setf p2 (xy (+ (xy-x p1) (xy-x s1) 6) (xy-y p1)))
      (setf s2 (string-extent "Dark Green" *capella-font* (xy 20 4)))
      (setf p3 (xy (+ (xy-x p2) (xy-x s2) 6) (xy-y p1)))
      (setf s3 (xy 40 18))
      (ccl:add-subviews
       view
       (make-instance 'label
         :title "Capella Font:" :view-position p1 :view-size s1)
       (make-instance 'pop-up-menu
         :view-position p2 :view-size s2
         :default-item (1+ (position (font-name *capella-font*)
                                     fonts :test #'string-equal))
         :menu-items (font-names fonts ':capella-font))
       (make-instance 'pop-up-menu
         :view-position p3 :view-size s3
         :default-item (1+ (position (font-size *capella-font*)  sizes))
         :menu-items (font-sizes sizes ':capella-font)))
      (setf p1 (xyinc p1 0 h))
      (setf p2 (xyinc p2 0 h))
      (setf p3 (xyinc p3 0 h))
      (ccl:add-subviews
       view
       (make-instance 'label :title "Listing Font:"
                      :view-position p1 :view-size s1)
       (make-instance 'pop-up-menu
         :view-position (xyinc p2 0 0) :view-size s2
         :default-item (1+ (position (font-name *listing-font*)
                                     ccl:*font-list* :test #'string-equal))
         :menu-items (font-names fonts ':listing-font))
       (make-instance 'pop-up-menu
         :view-position (xyinc p3 0 0) :view-size s3
         :default-item (1+ (position (font-size *listing-font*)  sizes))
         :menu-items (font-sizes sizes ':listing-font)))
      (setf p1 (xyinc p1 0 h))
      (setf p2 (xyinc p2 0 h))
      (setf p3 (xyinc p3 0 h))
      (let ((names '("White" "Light Gray" "Gray" "Dark Gray" "Black" "-"
                     "Pink" "Red" "Orange" "Yellow" "Green" "Dark Green"
                     "Light Blue" "Blue" "Purple" "Brown" "Tan" "-" 
                     "Other..."))
            (colors
             (list ccl:*white-color* ccl:*light-gray-color* ccl:*gray-color*
                   ccl:*dark-gray-color* ccl:*black-color* nil
                   ccl:*pink-color* ccl:*red-color* ccl:*orange-color*
                   ccl:*yellow-color* ccl:*green-color* ccl:*dark-green-color*
                   ccl:*light-blue-color* ccl:*blue-color* ccl:*purple-color*
                   ccl:*brown-color* ccl:*tan-color*)))
        (ccl:add-subviews
         view
         (make-instance 'label
           :title "Background Color:"
           :view-position p1 :view-size s1)
         (make-instance 'pop-up-menu
           :view-position (xyinc p2 0 0) :view-size s2
           :default-item (1+ (or (position *bgc* colors)
                                 (position "Other..." names
                                           :test #'string-equal)))
           :menu-items
           (nconc
            (loop for n in names as c in colors
                  collect
                  (let ((c c))
                    (make-instance 'ccl:menu-item
                      :menu-item-title n
                      :menu-item-action
                      (if c
                        #'(lambda (&aux (p (preferences-current view)))
                            (setf (getf p ':bgc) c))
                        nil))))
            (list
             (make-instance 'ccl:menu-item
                      :menu-item-title "-"
                      :disabled t
                      :menu-item-action nil)
             (make-instance 'ccl:menu-item
                      :menu-item-title "Other..."
                      :menu-item-action
                      #'(lambda ()
                          (let ((p (preferences-current view))
                                (nc (catch :cancel
                                      (ccl:user-pick-color :color *bgc*))))
                            (when nc
                              (setf (getf p ':bgc) nc))))))))))

      (setf p1 (xyinc p1 0 h))
      (setf p2 (xyinc p2 0 h))
      (setf p3 (xyinc p3 0 h))
      (setf w (+ 8 (xy-x p3) (xy-x s3) 8))
      (let ((s (string-extent "Load..." *capella-font* (xy 10 3))))
        (ccl:add-subviews
         view
         (make-instance 'button
           :title "Load..."
           :view-position (xy (- w 8 (* 3 (xy-x s)) (* 3 8)) (xy-y p3))
           :view-size s :enabled? nil)
        
         (make-instance 'button
           :title "Save..."
           :view-position (xy (- w 8 (* 2 (xy-x s)) (* 2 8)) (xy-y p3))
           :view-size s :enabled? nil)
        
         (make-instance 'default-button
           :view-position (xy (- w 8 (* 1 (xy-x s)) (* 1 8)) (xy-y p3))
           :title "Set"
           :action #'worksheet-action
           :view-nick-name ':Ok
           :view-size (xy+ s (xy 0 3)))))
        (ccl:set-view-size view (xy w (+ (xy-y p1) 30))))))
    
(defmethod initialize-worksheet ((view preferences))
  (setf (preferences-current view)
        (list ':capella-font
              (list (font-name *capella-font*) 
                    (font-size *capella-font*))
              ':listing-font 
              (list (font-name *listing-font*)
                    (font-size *listing-font*))
              ':bgc *bgc*)))
                               
(defmethod worksheet-action ((view preferences))
  (let ((current (slot-value view 'current)))
    (let ((pref (getf current ':capella-font)))
      (set-capella-font
       (make-font :name (first pref) :size (second pref)
                  :defaults *capella-font*)))
    (let ((pref (getf current ':listing-font)))
      (setf *listing-font* (make-font :name (first pref) 
                                      :size (second pref)
                                      :defaults *listing-font*)))
    (let ((pref (getf current ':bgc)))
      (setf *bgc* pref))
    (map-capella-windows #'ccl:invalidate-view)))


;;;
;;; (make-instance 'move-worksheet)
;;;

(defclass move-worksheet (framed-worksheet)
  ()
  (:default-initargs
    ))

(defmethod initialize-instance :after ((view move-worksheet)
                                       &rest ignore)
  (declare (ignore ignore))
  (let ((s (string-extent "Reverse" *capella-font* (xy 20 3)))
        (x 8)
        (y 8)
        (d 4))
    (ccl:add-subviews 
     view
     (make-instance 'radio-button
       :title "Rotate" :view-nick-name ':rotate
       :action #'move-radio-button-action
       :view-position (xy x y) :view-size s)
     (make-instance 'radio-button
       :title "Reverse" :view-nick-name ':reverse
       :action #'move-radio-button-action
       :view-position (xy (+ x (xy-x s) d) y) :view-size s))
    (setf y (+ y (xy-y s) 4))
    (ccl:add-subviews 
     view
     (make-instance 'radio-button
       :title "Swap" :view-nick-name ':swap
       :action #'move-radio-button-action
       :view-position (xy x y) :view-size s)
     (make-instance 'radio-button
       :title "Shuffle" :view-nick-name ':shuffle
       :action #'move-radio-button-action
       :view-position (xy (+ x (xy-x s) d) y) :view-size s)) 
    (setf y (+ y (xy-y s) 8))
    (let ((b (string-extent "OK" *capella-font* (xy 20 2))))
      (ccl:add-subviews 
       view
       (make-instance 'default-button
         :enabled? nil
         :view-position (xy (- (+ x (xy-x s) d) (floor (xy-x b) 2)) y)
         :view-size b
         :action #'worksheet-action
         :view-nick-name ':Ok))
      (setf y (+ y (xy-y b) 4 8)))
    (set-worksheet-title view "Move" *browser*)
    (ccl:set-view-size view (xy (+ x (* (xy-x s) 2) d x) y))))

(defun move-radio-button-action (b)
  (control-enable (ccl:find-named-sibling b ':ok)))

(defmethod initialize-worksheet ((view move-worksheet))
  ;; sheet is enabled if selection is range or 2 disjoint.
  (if (listing-browser? *browser*)
    (let* ((l (browser-listing *browser*))
           (n (sequence-nselected l)))
      (if (< n 2)
        (disable-all-dialogs view)
        (let ((b1 (ccl:view-named ':swap view))
              (b2 (ccl:view-named ':shuffle view))              
              (b3 (ccl:view-named ':rotate view))
              (b4 (ccl:view-named ':reverse view))
              (ok (ccl:view-named ':ok view)))
          (cond ((= n 2)
                 (control-enable b1)
                 (toggle-select b1))
                (t
                 (control-disable b1)
                 (toggle-deselect b1)))
          (cond ((contiguous-selection? l)
                 (control-enable b2)
                 (control-enable b3)
                 (control-enable b4))              
                (t
                 (control-disable b2)
                 (control-disable b3)
                 (control-disable b4)))
          (if (pushed-radio-button view)
            (control-enable ok)
            (control-disable ok))
          (set-worksheet-title view "Move" *browser*))))
    (disable-all-dialogs view)))

(defmethod worksheet-action ((view move-worksheet))
  (let ((listing (browser-listing *browser*)))
    (listing-move listing 
                  (ccl:view-nick-name (pushed-radio-button view)))))

;;;
;;; midi-play
;;;

#+MIDI
(progn

(defclass midi-play (framed-worksheet)
  ()
  (:default-initargs
    :window-title "Play"))

(defun install-play-file (w f)
  (let ((b (ccl:view-named ':playfile w)))
    (set-current-output-stream b f)
    (update-worksheet w)))

(defun open-play-file (w)
  (let ((file (ccl:choose-file-dialog :button-string "Play")))
    (when file
      (install-play-file
       w (open-event-stream 'midi-file ':initialize 
                            :name (namestring file))))))

(defun make-play-menu-item (f w)
  (make-instance 'object-menu-item
    :object f
    :menu-item-title (object-name f)
    :menu-item-action #'(lambda () (install-play-file w f))))

(defun make-play-menu (w pos siz)
  (make-instance 'ccl::pop-up-menu
    ;; mousing the popup fills menu with all current midi files.
    :update-function 
    #'(lambda (m)
        (update-streams-menu m #'(lambda (f) (make-play-menu-item f w))
                            'midi-file))
    :menu-items 
    (list 
     (make-instance 'ccl:menu-item :menu-item-title "Files" :disabled t)
     (make-instance 'ccl:menu-item :menu-item-title "-" :disabled t)
     (make-instance 'ccl:menu-item :menu-item-title "-" :disabled t)
     (make-instance 'ccl:menu-item :menu-item-title "Open..."
                    :menu-item-action #'(lambda () (open-play-file w))))
    :item-display "Files"
    :view-font *capella-font*
    :view-nick-name ':filemenu
    :view-position pos :view-size siz))

(defmethod initialize-instance :after ((view midi-play) &rest args)
  (declare (ignore args))
  (let ((x1 10) ; offset and spacing
        (y 10)
        (h 20)  ; line height
        (s1 (string-extent "Output " *capella-font*))
        (s2 (string-extent "Hard Disk:Users:hkt:midi:test.midi" 
                           *capella-font* ))
        x2 x3 
        (e1 (string-extent "From:" *capella-font* (xy 5 0)))
        (e2 (string-extent "To:" *capella-font* (xy 5 0)))
        (e3 (string-extent "Ahead:" *capella-font* (xy 5 0)))
        (e4 (string-extent "Zoom:" *capella-font* (xy 5 0)))
        e5 x1a x2a x3a
        menu fbuf 
        ; port modem printer
        froml from tol to hush play
        startl start zooml zoom box w)
    (setf x2 (+ x1 (xy-x s1) x1) )
    ;; column 3 is 1/2 way across file buffer.
    (setf x3 (+ (xy-x x2) (floor (xy-x s2) 2) ))
    (setf x1a (+ x1 (xy-x e3) 5))
    (setf x2a (+ x2 (xy-x e3) 5))
    (setf e5 (xy (-  (xy-x x3) (xy-x x2a) 5) (xy-y s1)))
    (setf x3a (+ x3 (xy-x e3) 5))
    (setf w (+ x1 (xy-x s1) x1 (xy-x s2) 13))
    (setf menu (make-play-menu view (xy x1 y) 
                               (xy (+ x1 (xy-x s1) (xy-x s2))
                                   (+ 5 (xy-y s2)))))
    (incf y (+ h 4))
    (setf fbuf (make-instance 'output-stream-edit-button
                 :view-nick-name ':playfile 
                 :view-position (xy x1 y) 
                 :view-size (xy (+ x1 (xy-x s1) (xy-x s2))
                                (+ 5 (xy-y s2)))))
    (incf y (+ h 4))
    (setf froml (make-instance 'label 
                  :title "From:"
                  :view-position (xy x1 y) :view-size e1))
    (setf from (make-instance 'textbuf 
                 :view-nick-name ':from
                 :view-position (xy x1a y) :view-size e5))
    (setf tol (make-instance 'label 
                :title "To:"   
                :view-position (xy x3 y) :view-size e2))
    (setf to (make-instance 'textbuf 
               :view-nick-name ':to
               :view-position (xy x3a y) :view-size e5))
    (incf y (+ h 4))
    (setf startl (make-instance 'label 
                   :title "Ahead:"
                   :view-position (xy x1 y)
                   :view-size e3))
    (setf start (make-instance 'textbuf :view-nick-name ':start
                               :view-position (xy x1a y) :view-size e5))
    (setf zooml (make-instance 'label :title "Zoom:"   
                               :view-position (xy x3 y) :view-size e4))
    (setf zoom (make-instance 'textbuf :view-nick-name ':zoom
                              :view-position (xy x3a y) :view-size e5))
    (incf y h)
    (setf box (make-instance 'check-box
                 :title "Async. playback" :view-nick-name ':async
                 :view-position (xy x1 y)
                 :action #'(lambda (b) b (midi-hush))))
    (setf hush (make-instance 'button
                 :title "Hush" :view-nick-name ':Hush
                 :view-position (xy (floor w 2) (+ y 2))
                 :action #'(lambda (b) b (midi-hush))))
    (setf play (make-instance 'default-button
                 :title "Play"
                 :view-nick-name ':play
                 ; :view-position (xy x3a y)
                 :view-position (xy (- x3a 5) (- y 2))
                 :action #'worksheet-action))
    (ccl:add-subviews
     view menu fbuf
     ; port modem printer 
     froml from tol to
     startl start zooml zoom box hush play)
    (ccl:set-view-size view (xy w (+ y h 5)))))

(defmethod initialize-worksheet ((view midi-play))
   (update-worksheet view))

(defmethod update-worksheet ((view midi-play))
  (let ((file (current-output-stream
               (ccl:view-named ':playfile view)))
        fileok)
    (if (setf fileok (if file (probe-file (object-name file)) nil))
      (progn
        (control-enable (ccl:view-named ':from view))
        (control-enable (ccl:view-named ':to view)))
      (progn
        (control-disable (ccl:view-named ':from view))
        (control-disable (ccl:view-named ':to view))))
    (if (and fileok (midi-open?))
      (progn
        (control-enable (ccl:view-named ':hush view))
        (control-enable (ccl:view-named ':play view)))
      (progn
        (control-disable (ccl:view-named ':hush view))
        (control-disable (ccl:view-named ':play view))))))

(defmethod worksheet-action ((view midi-play))
  (let ((file (current-output-stream
               (ccl:view-named ':playfile view)))
        (form nil)
        (args )) ; nreversed later
    (if (and file (probe-file (object-name file)))
      (progn 
        (push (object-name file) args)
        (let ((map (slot-value-or-default file 'chanmap)))
          (when map
            (push ':chanmap args)
            (push map args))))
      (return-from worksheet-action
        (warn-user "File: ~S does not exist." form)))
    (push ':spool? args)
    (push  (not (stella-p)) args)
    (when (setf form (textbuf-form (ccl:view-named ':from view)))
      (if (and (numberp form) (>= form 0))
        (progn (push ':start args) (push form args))
        (return-from worksheet-action
          (warn-user "From: ~S not >= 0" form))))
    (when (setf form (textbuf-form (ccl:view-named ':to view)))
      (if (and (numberp form) (>= form 0))
        (progn (push ':end args) (push form args))
        (return-from worksheet-action
          (warn-user "To: ~S not >= 0" form))))
    (when (setf form (textbuf-form (ccl:view-named ':start view)))
      (if (and (numberp form) (>= form 0))
        (progn (push ':headstart args) (push form args))
        (return-from worksheet-action 
          (warn-user "Start: ~S not >= 0" form))))
    (when (setf form (textbuf-form (ccl:view-named ':zoom view)))
      (if (and (numberp form) (>= form 0))
        (unless (= form 1)
          (progn (push ':timescale args) (push form args)))
        (return-from worksheet-action
          (warn-user "Start: ~S not >= 0" form))))
    (when (toggle-selected? (ccl:view-named ':async view))   
      (push ':async? args) (push t args))
    (apply #'midifile-play (nreverse args))))

;;;
;;; Midi Receive
;;;

(defclass midi-receive (framed-worksheet)
  ((initializer :initform nil)
   (finalizer :initform nil)
   (receiver :initform nil)
   (stop-mode :initform ':mouse)
   (reset-time? :initform nil)
   )
  (:default-initargs
    :window-title "Receive"
    :view-font *capella-font* ))

(defmethod initialize-instance :after ((view midi-receive) &rest args)
  (declare (ignore args))
  (let* ((x1 10)
         (y 10)
         (fh (font-height *capella-font*))
         (lh (+ 10 fh))
         (lw (xy-x (string-extent "Initializer:-" *capella-font*)))
         (bw (xy-x (string-extent "MIDI-RECORD-MESSAGE" *capella-font*)))
         (x2 (+ x1 lw ))
         width button buttons
         lab1 lab2 lab3 receiver initializer finalizer )
    
    (setf buttons
          (loop with x = 10 and space = 10
                for name in '(:Through :Record :Print :Receive)
                for radio = (make-instance 'radio-button
                              :view-position (xy x y)
                              :selected? (eq name ':through)
                              :title (string-capitalize (symbol-name name))
                              :action #'update-worksheet
                              :view-nick-name name)
                collect radio
                do (setf x (+ x (view-width radio) space))
                finally (setf width x)))
    (apply #'ccl:add-subviews view buttons)
    
    (incf y lh)
    (setf lab1 (make-instance 'label :view-position (xy x1 y)
                              :title "Receiver:"))
    (setf receiver
          (make-instance 'textbuf
            :dialog-item-enabled-p nil
            :view-position (xy x2 y)
            :view-size (xy bw fh)
            :title "" :action nil :view-nick-name ':receiver))
    
    (incf y lh)
    (setf lab2 (make-instance 'label :view-position (xy x1 y)
                              :title "Initializer:"))
    (setf initializer
          (make-instance 'textbuf 
            :view-position (xy x2 y)
            :view-size (xy bw fh)
            :title "" :action nil :view-nick-name ':initializer))
    
    (incf y lh)
    (setf lab3 (make-instance 'label :view-position (xy x1 y)
                              :title "Finalizer:"))
    (setf finalizer
          (make-instance 'textbuf 
            :view-position (xy x2 y)
            :view-size (xy bw fh)
            :title "" :action nil :view-nick-name ':finalizer))
    
    (incf y lh)
    (setf button
          (make-instance 'default-button 
            :title "Start"
            :action #'worksheet-action
            :view-nick-name ':button))
    (ccl:add-subviews
     view lab1 receiver lab2 initializer lab3 finalizer button )
    
    (ccl:set-view-size view (xy (max width (+ x2 bw 10))
                                (+ y lh 10)))
    (ccl:set-view-position button
                           (xy (- (floor (view-width view) 2)
                                  (floor (view-width button) 2))
                               y))))

(defmethod initialize-worksheet ((view midi-receive))
  (update-worksheet view))

(defmethod update-worksheet ((view midi-receive))
  (let ((user (ccl:view-named ':receive view))
        (receiver (ccl:view-named ':receiver view)))
    (if (toggle-selected? user)
      (ccl:dialog-item-enable receiver)
      (ccl:dialog-item-disable receiver))
    (unless (stella-p )
      (warn-user 
       "Receiving MIDI without Stella running may cause dropouts."))))

(defmethod worksheet-action ((view midi-receive))
  (let ((button (ccl:view-named ':button view)))
    (if (string-equal (item-title button) "Stop")
      (progn (midi-stop-receiving)
             (set-item-title button "Start"))
      (flet ((insure-function (buf)
               (let ((fn (textbuf-function buf)))
                 (if fn fn
                     (progn
                       (warn-user "~S cannot be interpreted as a function."
                                  (textbuf-text buf))
                       (return-from worksheet-action nil))))))
        (let (recv args)
          (cond ((toggle-selected? (ccl:view-named ':through view))
                 (setf recv #'midi-write-message))
                ((toggle-selected? (ccl:view-named ':record view))
                 (setf recv (make-recording-closure)))
                ((toggle-selected? (ccl:view-named ':print view))
                 (setf recv #'midi-print-message))
                ((toggle-selected? (ccl:view-named ':receive view))
                 (setf recv (insure-function (ccl:view-named ':receiver view)))))
          (unless (nullstringp 
                   (textbuf-text (ccl:view-named ':initializer view)))
            (push ':initializer args)
            (push (insure-function (ccl:view-named ':initializer view))
                  args))
          (unless (nullstringp (textbuf-text 
                                (ccl:view-named ':finalizer view)))
            (push ':finalizer args)
            (push (insure-function (ccl:view-named ':finalizer view))
                  args))
          (apply #'midi-receive recv (nreverse args))
          (set-item-title button "Stop"))))))

  
) ; #+MIDI


