;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :common-music)

(defun %map-into (y scale offset return-type &aux spec)
  (when scale (setf y (* y scale)))
  (when offset (setf y (+ y offset)))
  (if return-type
    (progn
      (when (consp return-type) 
        (setf spec (second return-type)
              return-type (first return-type)))
      (case return-type
        ((integer fixnum) (round y))
        ((float double-float single-float long-float) 
         (if spec 
             (let ((power (expt 10.0 spec)))
               (/ (round (* y power)) power))
           (float y)))
        ((ratio rational) (rationalize y))
        (t (error "Return type ~S not implemented." return-type))))
    y))

(defun rescale (value min max newmin newmax &key return-type)
  (declare (inline %map-into))
  (let ((v (+ (* (/ (- newmax newmin) (- max min)) (- value min))
              newmin)))
    (if return-type 
      (%map-into v nil nil return-type)
      v)))      

#|
(defun rescale! (value min max newmin newmax &key return-type)
  (declare (inline %map-into)
           (number  min max newmin newmax ))
  (let* (
         ;(nr (- newmax newmin))
         ;(or (- max min))
         ;(nv (- value min))
         ;(ra (/ nr or))
         ;(sv (* ra nv))
         ;(v (+ sv newmin))
         tmp1 tmp2 )
    (declare (number tmp1 tmp2 ;nr or nv v ra sv
                     )
             (dynamic-extent tmp1 tmp2 ; nr or nv ra
                             )
             (optimize (speed 3) (safety 0)))
    (setf tmp1 (- newmax newmin))
    (setf tmp2 (- max min))
    (setf tmp1  (/ tmp1 tmp2))
    (setf tmp2 (- value min))
    (setf tmp1 (* tmp1 tmp2))
    (if return-type (%map-into (+ tmp1 newmin) nil nil return-type)
        (+ tmp1 newmin))))
|#

(defconstant 2pi (* 2 pi))

(defmacro with-vars-snapshotted ((&rest vars) &body body)
  `(let ,(mapcar #'(lambda (v) (list v v)) vars) ,@body))

(defun between (lb ub &optional exception (state *cm-state*))
  (let ((range (- ub lb)))
    (if (not (> range 0))
	lb
      (if exception
	  (loop with num 
	   when (/= exception (setf num (+ lb (random range state))))
	   return num)
	(+ lb (random range state))))))

(defun pickl (&rest args)
  (elt args (random (length args) *cm-state*)))

(defun pick (seq &key (end (length seq)) (start 0)
                 avoid (state *cm-state*))
  (let ((range (- end start)))
    (if avoid
      (loop with x = avoid
            while (eq x avoid)
            do (setf x (elt seq (+ start (random range state))))
            finally (return x))
      (elt seq (+ start (random range *cm-state*))))))


(defparameter *coordinates-are-x-y-pairs* T
  "If true, coordinates are parsed as X Y pairs")

(defun interpolation (&rest coords)
  (unless (and coords (evenp (length coords)))
    (error "Malformed interpolation list: ~S" coords))
  (let ((l (if *coordinates-are-x-y-pairs*
	       coords
	     (loop for (y x) on coords by #'cddr nconc (list x y)))))
    (loop with x2 for x1 in l by #'cddr
          unless (or (null x2)(< x2 x1))
          do (error "X value ~S not < ~S in ~S" x2 x1 l)
          do (setf x2 x1))
    l))

(defun interp1 (x coords)
  (let ((sav coords))
    (macrolet ((getcoord (l)
		 `(or (pop ,l)
		      (error "Found odd number of coordinates: ~S"
                             sav))))
      (let* ((x1 (getcoord coords))
             (y1 (getcoord coords))
             (x2 x1)
             (y2 y1))
        (loop when (or (null coords)
                       (> x2 x))
              return nil
              do
              (setf x1 x2 y1 y2 
                    x2 (getcoord coords) y2 (getcoord coords)))
        (cond ((>= x x2)
               y2)
              ((<= x x1)
               y1)
              (t
               (+ y1 (* (- x x1) (/ (- y2 y1) (- x2 x1))))))))))

(defun interpl (x &rest coords)
  (declare (inline interp1))
  (let ((y (interp1 x coords)))
    (%map-into y nil nil nil)))

(defun interp (x coords &key scale offset return-type)
  (declare (inline interp1 %map-into))
  (%map-into (interp1 x coords) scale offset return-type))

(defun tendency (x low high &key scale offset
                   return-type (random #'random)
                   (state *cm-state*) &aux value)
  (declare (inline %map-into))
  (when (consp low) (setf low (interp1 x low)))
  (when (consp high) (setf high (interp1 x high)))
  (when (> low high) (rotatef low high))
  (setf value (if (= low high)
                  (float low)
                  (+ low (funcall random (float (- high low) ) state))))
  (%map-into value scale offset (or return-type 'float)))


#|
(defun interp (x coords &key scale offset return-type)
  (declare (inline interp1))
  (let ((y (interp1 x coords)))
    (when scale (setf y (* y scale)))
    (when offset (setf y (+ y offset)))
    (when return-type
      (setf y (if (eq return-type 'integer) (round y)
                  (unless (eq return-type 'float) 
                    (error "Return type ~S not implemented." 
                           return-type)))))
    y))

(defun tendency (x low high &key scale offset return-type 
                   (random #'random)
                   (state *cm-state*))
  (when (consp low) (setf low (interp1 x low)))
  (when (consp high) (setf high (interp1 x high)))
  (when (> low high) (rotatef low high))
  (let ((y (if (= low high) low
               (+ low (funcall random (- high low) state)))))
    (when scale (setf y (* y scale)))
    (when offset (setf y (+ y offset)))
    (when return-type
      (setf y (if (eq return-type 'integer) (round y)
                  (unless (eq return-type 'float) 
                    (error "Return type ~S not implemented." 
                           return-type)))))
    y))
|#

(defun function-value (x coords &optional scale offset)
  ;; leave around for backward compatability
  (warn "FUNCTION-VALUE has been renamed INTERP. ~
         Please change your code.")
  (interp x coords  :scale scale :offset offset))

(defun lookup (x coords)
  ;; no interpolation
  (macrolet ((getcoord (l &optional odd?)
               `(let ((v (pop ,l)))
                  (when (and ,odd? (null ,l) )
                    (error "Found odd number of coordinates: ~S"
                           sav))
                  v)))
    (let* ((sav coords)
	   (x1 (getcoord coords t))
	   (y1 (getcoord coords))
	   (x2 x1)
	   (y2 y1))
      (loop when (or (null coords) (> x2 x))
            return nil
            do
	    (setq x1 x2 y1 y2
	          x2 (getcoord coords t) y2 (getcoord coords)))
      (if (>= x x2) y2 y1))))




;;;
;;; (setf p (pfunc '(0 0 .1 0 .15 1 .25 0 .75 0 .8 1 1 0)))
;;; (dotimes (i 100) (format t "~%~$" (pfunc (random 1.0) p)))
;;;

(defun pfunc (probs)
  (let ((sum 0.0) 
        (len (length probs))
        areas)
    (unless (and (evenp len) (> len 2))
      (error "Bad coordinate list: ~S" probs))
    (setf areas
          (loop for ctr from 2 below len by 2
                for y1 = (elt probs (- ctr 1))
                for y2 = (elt probs (+ ctr 1))
                for x1 = (elt probs (- ctr 2))
                for x2 = (elt probs ctr)
                for area = (* (- x2 x1) (+ y1 (/ (- y2 y1) 2)))
                do (incf sum area) 
                collect area))
    (loop with l = () and x = 0.0
          for a in areas
          for dx = (/ a sum)
          for y = 0 then (+ y 2)
          do
          (when (> dx 0.0)
            (push x l)
            (push (elt probs y) l)
            (incf x dx)
            (push x l)
            (push (elt probs (+ 2 y)) l))
          finally (return (nreverse l)))))

(defun fit (number lb ub &optional (mode :reflect) &key return-type)
  (declare (inline %map-into))
  (when (> lb ub) (rotatef lb ub))
  (%map-into
    (if (<= lb number ub) number
        (let ((b (if (> number ub) ub lb))
              (r (- ub lb)))
          (ecase mode
            (:reflect
             (let* ((2r (* 2 r))
                    (v (rem (- number b) 2r)))
               (+ (if (> (abs v) r)
                    (funcall (if (>= v 0) #'- #'+) v 2r)
                    (- v))
                  b)))
            (:wrap
             (+ (if (= b ub) lb ub)
                (rem (- number b) r))))))
    nil nil return-type))

(defun expl (power &key (y0 0.0) (y1 1.0) (base 2) return-type)
  (declare (inline %map-into))
  (unless (<= 0 power 1) (error "Power ~S is not between 0 and 1." power))
  (%map-into
    (if (= base 1)
        (+ y0 (* (- y1 y0) power))
      (+ y0 (* (/ (- y1 y0) (- base 1.0)) (- (expt base power) 1.0))))
    nil nil return-type))

;;;
;;; explsegs returns a list of exponentially scaled segments that sum to
;;; the specified number. Num is the number to sum to.  Len is the number
;;; of segments to sum over.  Power is the power curve for the
;;; exponential scaling.  power numbers 0<n<1 yield concave slopes.
;;;

(defun explsegs (len sum power &key return-type)
  (declare (inline %map-into))
  (if (= len 0) (* 0 sum)  ; return 0 in same type as num
      (loop for i from 1 to len
            for l = 0 then v
            for v = (expl (/ i len) :base power)
            collect (%map-into (* sum (- v l))
                               nil nil return-type))))

;;;
;;; like explsegs but returns each value individually
;;;

(defun explseg (i len num power &key return-type)
  (declare (inline %map-into))
  (if (>= i len) (* 0 num) ; return 0 in same type as nu,
      (%map-into
       (* num (- (expl (/ (1+ i) len) :base power)
                 (if (<= i 0) 0 (expl (/ i len) :base power))))
       nil nil return-type)))

(defun shuffle (seq &key (start 0) (end (length seq) )
                    (state *cm-state*) (copy t)
                    &aux (width (- end start)))
  (if (< width 2)
    seq
    (progn
      (when (and copy (typep seq 'list))
        (setf seq (copy-list seq)))
      (loop for i from start to (1- end)
            do 
            (rotatef (elt seq i) 
                     (elt seq (+ start 
                                 (random width state)))))
      seq)))






