;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :stella)

;;;
;;; this file contains low level routines for parsing and processing
;;; command input data.
;;;

(defun balanced-form? (string &optional (start 0)
                              (end (length string)))
  ;; true if string contains balanced lisp form.
  (loop with lev = 0
        and char and done  
        while (< start end) do
        (setf char (elt string start))
        (cond ((char= char #\()
               (incf lev))
              ((char= char #\))
               (when (= (decf lev) 0) 
                 (setf done t start (1+ start))))
              ((char=  char #\Space) 
               (when (= lev 0) (setf done t))))
        until done
        do (incf start)
        finally (return (and (= lev 0) (= start end)))))
;;;
;;;
;;;

(defun make-lambda (form) (coerce `(lambda () ,form) 'function))

;;;
;;; argfield finds the limits of the next argument in an input string, taking
;;; care to parse object references, pathnames, sequences and lisp forms
;;; correctly. an argument may include whitespace if it uses type delimitors:
;;; :pathname  "", :reference [], :form (), :sequence ,.
;;;

(defun argfield (string &optional (start 0) end type)
  (unless type (setf type ':form))
  (let* ((beg start)
         (end (or end (length string)))
         (lev 0)
         (bag '(#\space #\tab #\,))
         (white '(#\space #\tab))
         (seq (or (eq type ':sequence)
                  (eq type ':reference)))
         #-(or kcl aclpc) (char #\null)
         #+kcl (char (code-char 0))
         #+aclpc (char #\return)     ; no #\null or (char-code 0) in aclpc!
         pos next done
         sin )
    (loop while (< beg end)
          while (member (setf char (elt string beg))
                        white)
          do (incf beg))
    (if (char= char #\,)
      (error "Illegal ,")
      (setf pos beg))
     (case type
      ((:reference :range :position)
       (setf seq (if (eq type :reference) t nil))
       (loop while (< pos end) do 
             (setf char (elt string pos))
             (cond ((char= char #\[)
                    (incf lev))
                   ((char= char #\])
                    (if (= (decf lev) 0)
                      (setf done t pos (1+ pos))))
                   ((and (= lev 0) (member char bag))
                    (setf done t)))
             until done
             do (incf pos))
       (unless (= lev 0)
         (error "Unbalanced ~:[]~;[~]" (> lev 0))))
      (:pathname
       (loop while (< pos end) do
             (setf char (elt string pos))
             (cond ((char= char #\")
                    (if (= pos beg)
                      (incf lev)
                      (if (= (decf lev) 0)
                        (setf done t pos (1+ pos)))))
                   ((and (= lev 0) (member char bag))
                    (setf done t)))
             until done
             do (incf pos))
       (unless (= lev 0) (error "Unbalanced \"")))
      ((:form :symbol :keyword :sequence :container :containers t) 
       (setf seq (if (member type '(:sequence :containers)) t nil))
       (loop while (< pos end) do
             (setf char (elt string pos))
             (cond ((char= char #\()
                    (if (or (= pos beg) (> lev 0))
                      (incf lev)
                      (setf done t)))
                   ((char= char #\))
                    (if (= (decf lev) 0)
                      (setf done t pos (1+ pos))))
                   ((char= char #\")
                    (if sin      ; inside of string
                      (if (= (decf lev) 0)
                        (setf done t pos (1+ pos)))
                      (incf lev)
                      )
                    (setf sin (not sin))
                    ;(if (= pos beg)
                    ;  (incf lev)
                    ;  (if (= (decf lev) 0)
                    ;    (setf done t pos (1+ pos))))
                    )
                   ((and (= lev 0)
                         (member char bag))
                    (setf done t)))
             until done
             do (incf pos))
       (unless (= lev 0)
         (error "Unbalanced ~:[)~;(~]" (> lev 0))))
      ((:pairs :raw :rest)
       (setf pos 
             (1+ (position-if-not 
                  #'(lambda (x) (member x bag))
                  string :start beg :from-end t))))
      (t 
       (error "Don't know how to parse ~S args." type)))
    ;; pos is now 1 past the arg. find the pos
    ;; of the next arg, if any, and recurse if
    ;; we hit a comma and sequences are allowed.
    (unless (= pos end)
      (let ((n pos))
        (loop while (< n end)
              while (member (setf char (elt string n))
                            white)
              do (incf n))
        (when (< n end)
          (setf next n)
          (when (char= char #\,)
            (if seq
              (multiple-value-bind 
                (x y z) 
                (argfield string (1+ next) end type)
                (unless x (error "Dangling ,"))
                (setf pos y next z))
              (error "Illegal ,"))))))    
    (if (= beg end)
      (values nil nil nil)
      (values beg pos next))))

;;;
;;; with-args parses an input string according to the specified specs.
;;; each spec is (var &optional type init) where var is a variable that
;;; receives the parsed argument, type is the argument's type and
;;; init is an initial value for var.  There are currenty 8 types:
;;;    t (default)    a string
;;;    :form          lisp form 
;;;    :pathname      a file name, can include spaces if delimited by ""
;;;    :reference     a reference, can include spaces if delimited by []
;;;    :sequence      a series of elements delimited by ,
;;;    :rest          the rest of the args returned as list of :forms
;;;    :pairs         the rest of the args returned as list of {symbol <form>}
;;;    :keys          the rest of the args returned as list of {:symbol <form>}
;;;    :raw           the rest of the input returned as string
;;;    

(defmacro with-args (input (&rest specs) &body body)
  (let ((svar (gensym))
        (bvar (gensym))
        (evar (gensym))
        (nvar (gensym))
        (lvar (gensym))
        (argnum (length specs))
        string argerr user types syntax nullok)
    (if (consp input)
        (setf string (car input)
              argerr (if (getf (cdr input) ':argchecking)
                         (format nil "Command input ~~S contains more than ~A ~
                                 argument~@[s~]."
                                 argnum (> argnum 1)))
              syntax (getf (cdr input) ':syntax)
              nullok (getf (cdr input) ':nullok))
      (setf string input))
    (when (and argerr syntax)
      (setf argerr
            (concatenate 'string argerr
                         (format nil "~&Syntax: ~A" syntax))))
    (loop with var and type and init
          for spec in specs
          when (consp spec)
          do 
      (setf var (car spec) type (second spec) init (third spec))
          else do
      (setf var spec type nil init nil)
          collect (list var init) into vlist
          collect type into tlist
          finally (setf user vlist types tlist))
    `(let* ((,svar (or ,string ""))
            (,bvar 0)
            (,lvar (length ,svar))
            ,evar ,nvar ,@ user )
       (block parse-block
         ,@(if nullok
             `((when (or (null ,svar) (nullstringp ,svar))
               (return-from parse-block nil))))
         ,@(if syntax
             `((when (and (not *command-prompting*)
                          (or (null ,svar) (nullstringp ,svar)))
                 (tell-user ,syntax)
                 (cmdreturn nil))))
         ,.(loop for binding in user for type in types
                 for n from 1
            if (eq type :raw)
            collect `(setf ,(first binding) (subseq ,svar ,bvar))
            else if (or (eq type :rest) (eq type :pairs) (eq type :keys))
            collect 
              (let ((local (gensym))
                    (results (gensym))
                    (pairerr (format nil "Uneven pair arguments in: ~~S.~@[~~&Syntax: ~A~]"
                                     syntax)))
                `(let (,local ,results)
                   (loop while ,bvar
                         do
                     ,(parse-clause 
                        svar local bvar evar nvar lvar 
                        (if (eq type :rest) :form 
                            (if (eq type :keys) :keyword :symbol)
                            )
                        (if (eq type :rest) nil
                         `(cmderror ,pairerr ,svar) )
                        nil)
                     (push ,local ,results)
                    ,@(and (or (eq type :pairs) (eq type :keys))
                           (list (parse-clause svar local bvar evar nvar lvar 
                                              :form t nil)
                                 `(push ,local ,results)))

                     )
                   (setf ,(first binding) (nreverse ,results))))
            else
            collect (parse-clause svar (first binding)
                                  bvar evar nvar lvar type
                                  '(return-from parse-block nil)
                                  (and argerr (= n argnum)
                                       `(cmderror ,argerr ,svar))
                                  )))
       ,@body)))

(defun parse-clause (strvar valvar begvar endvar nxtvar lenvar type
                            if-no-next if-next)
  `(progn
     (multiple-value-setq (,begvar ,endvar ,nxtvar)
       (argfield ,strvar ,begvar ,lenvar , type))  
     (when ,begvar
       ,(cond ((or (eq type ':symbol) (eq type ':keyword) (eq type ':form))
               (let ((clause
                      (if (eq type ':form)
                        `(read-from-string ,strvar nil :eof :start ,begvar :end ,endvar)
                        `(intern (string-upcase (subseq ,strvar ,begvar ,endvar))
                                 ,(if (eq type :keyword) ':keyword '*package*)))))
                 `(setf ,valvar ,clause)))
              
              (t
               `(setf ,valvar (subseq ,strvar ,begvar ,endvar))))
       (setf ,begvar ,nxtvar))
     ,@ (if if-no-next `((unless ,begvar , if-no-next)) nil)
        ,@ (if if-next `((when ,begvar , if-next)) nil)
           ))


(defun split-input (string &optional type)
  (multiple-value-bind (b e n) (argfield string 0 nil type)
      (if b (values (subseq string b e)
                    (if n (subseq string n) "")))))

(defun sequence-to-list (string &key (beg 0) (end (length string)) 
                                     (read t))
  ;; the sequence elements must be delimited by comma
  (setf beg (tl:next-token-start string :start beg :end end))
  (let ((e end))
    (loop while beg
          do
         (setf e (or (tl:next-whitespace-start string :start beg :end end
                                              :whitespace '(#\,))
                       end))
          collect (if read 
                      (multiple-value-bind (x ee)
                        (read-from-string string nil :eof :start beg :end e)
                        (when (< ee e) 
                          (return-from sequence-to-list (values nil nil)))
                        x)
                     (let ((p e))
                       (loop while (char= (elt string (1- p)) #\space)
                             do (decf p))
                       (subseq string beg p))) 
          into results until (= e end)
          do (setf beg
               (tl:next-token-start string :start e :end end
                                   :whitespace '(#\, #\space #\tab)))
          finally (return (values results (= e end))))))


#|
(with-args "hoo bar z" ((x :reference) 
                                (y :pathname "test.snd")
                                z 
                                (a nil 123))
  (format t "Args: x=~S  y=~S  z=~S  a=~S"  x y z a))

  
(with-args "foo[1,2,4] 0 port a syntax midi"
  ((r :reference) 
   (o )
   (l :rest))
  (format t "Args: r=~S  o=~S  l=~S"  r o l))  
|#

;;;
;;; command functions
;;;

(defmacro defcmd (name lambda &body body)
  (let (comment declare)
    (when (stringp (first body))
      (setf comment (list (pop body))))
    (when (eq (first (first body)) 'declare)
       (setf declare (list (pop body))))
    `(defun ,name ,lambda
       ,@comment
       ,@declare
       (block command-function ,@body))))

(defmacro cmderror (&rest warnforms)
  (if warnforms
      `(progn 
         (warn-user ,@warnforms)
         (return-from command-function ':error))
    `(return-from command-function ':error)))

(defmacro cmdreturn (&optional value)
  `(return-from command-function ,value))

(defmacro cmdabort ()
  '(cmdreturn ':aborted))

(defmacro cmdabort? (form)
  `(if (eq ,form ':aborted) (cmdabort)))

(defmacro insure-arg (var &body form)
  (let ((ivar (gensym))
        (var1 (gensym))
        (var2 (gensym))
        (inputv nil))
    (when (consp var)
      (setf inputv (getf ':input var)
            var (car var)))
    `(let ((,ivar ,(or inputv var)))
       (multiple-value-bind (,var1 ,var2) 
                            (,@(car form) :input (or ,ivar "")
                                    :error-return (not *command-prompting* ))
         ,var2   ; not using for now.
         (when (find ,var1 '(:aborted :error))
           (return-from command-function ,var1))
         (setf ,var ,var1)))))



